/* PulsePosition Library for Teensy 3.1
 * High resolution input and output of PPM encoded signals
 * http://www.pjrc.com/teensy/td_libs_PulsePosition.html
 * Copyright (c) 2014, Paul Stoffregen, paul@pjrc.com
 *
 * Development of this library was funded by PJRC.COM, LLC by sales of Teensy
 * boards.  Please support PJRC's efforts to develop open source software by
 * purchasing Teensy or other PJRC products.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice, development funding notice, and this permission
 * notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "Arduino.h"

#include "PulsePosition_t4.h"


// Timing parameters, in microseconds.


// The shortest time allowed between any 2 rising edges.  This should be at
// least double TX_PULSE_WIDTH.
#define TX_MINIMUM_SIGNAL   300.0

// The longest time allowed between any 2 rising edges for a normal signal.
#define TX_MAXIMUM_SIGNAL  2500.0

// The default signal to send if nothing has been written.
#define TX_DEFAULT_SIGNAL  1500.0

// When transmitting with a single pin, the minimum space signal that marks
// the end of a frame.  Single wire receivers recognize the end of a frame
// by looking for a gap longer than the maximum data size.  When viewing the
// waveform on an oscilloscope, set the trigger "holdoff" time to slightly
// less than TX_MINIMUM_SPACE, for the most reliable display.  This parameter
// is not used when transmitting with 2 pins.
#define TX_MINIMUM_SPACE   5000.0

// The minimum total frame size.  Some servo motors or other devices may not
// work with pulses the repeat more often than 50 Hz.  To allow transmission
// as fast as possible, set this to the same as TX_MINIMUM_SIGNAL.
#define TX_MINIMUM_FRAME  20000.0

// The length of all transmitted pulses.  This must be longer than the worst
// case interrupt latency, which depends on how long any other library may
// disable interrupts.  This must also be no more than half TX_MINIMUM_SIGNAL.
// Most libraries disable interrupts for no more than a few microseconds.
// The OneWire library is a notable exception, so this may need to be lengthened
// if a library that imposes unusual interrupt latency is in use.
#define TX_PULSE_WIDTH      100.0

// When receiving, any time between rising edges longer than this will be
// treated as the end-of-frame marker.
#define RX_MINIMUM_SPACE   3500.0

// convert from microseconds to I/O clock ticks
#define CLOCKS_PER_MICROSECOND (150./4)  // pcs 8+2
#define TX_MINIMUM_SPACE_CLOCKS   (uint32_t)(TX_MINIMUM_SPACE * CLOCKS_PER_MICROSECOND)
#define TX_MINIMUM_FRAME_CLOCKS   (uint32_t)(TX_MINIMUM_FRAME * CLOCKS_PER_MICROSECOND)
#define TX_PULSE_WIDTH_CLOCKS     (uint32_t)(TX_PULSE_WIDTH * CLOCKS_PER_MICROSECOND)
#define TX_DEFAULT_SIGNAL_CLOCKS  (uint32_t)(TX_DEFAULT_SIGNAL * CLOCKS_PER_MICROSECOND)
#define RX_MINIMUM_SPACE_CLOCKS   (uint32_t)(RX_MINIMUM_SPACE * CLOCKS_PER_MICROSECOND)

//Leave Frame stuff for now
//
//
// Output 
#define CTRL_SET TMR_CTRL_CM(1) | TMR_CTRL_PCS(8 + 2) | TMR_CTRL_LENGTH |TMR_CTRL_OUTMODE(2)
#define CTRL_CLEAR TMR_CTRL_CM(1) | TMR_CTRL_PCS(8 + 2) | TMR_CTRL_LENGTH |TMR_CTRL_OUTMODE(1)
#define PULSEPOSITION_MAXCHANNELS 16


const PulsePositionOutput::TMR_Hardware_t PulsePositionOutput::hardware[] = {
	{ 6,1, &IMXRT_TMR4, &CCM_CCGR6, CCM_CCGR6_QTIMER4(CCM_CCGR_ON), IRQ_QTIMER4, &PulsePositionOutput::isrTimer4 },
	{ 9,2, &IMXRT_TMR4, &CCM_CCGR6, CCM_CCGR6_QTIMER4(CCM_CCGR_ON), IRQ_QTIMER4, &PulsePositionOutput::isrTimer4 },
    {10,0, &IMXRT_TMR1, &CCM_CCGR6, CCM_CCGR6_QTIMER1(CCM_CCGR_ON), IRQ_QTIMER1, &PulsePositionOutput::isrTimer1 },
    {11,2, &IMXRT_TMR1, &CCM_CCGR6, CCM_CCGR6_QTIMER1(CCM_CCGR_ON), IRQ_QTIMER1, &PulsePositionOutput::isrTimer1 },
    {12,1, &IMXRT_TMR1, &CCM_CCGR6, CCM_CCGR6_QTIMER1(CCM_CCGR_ON), IRQ_QTIMER1, &PulsePositionOutput::isrTimer1 },
    {13,0, &IMXRT_TMR2, &CCM_CCGR6, CCM_CCGR6_QTIMER2(CCM_CCGR_ON), IRQ_QTIMER2, &PulsePositionOutput::isrTimer2 },
    {14,2, &IMXRT_TMR3, &CCM_CCGR6, CCM_CCGR6_QTIMER3(CCM_CCGR_ON), IRQ_QTIMER3, &PulsePositionOutput::isrTimer3 },
    {15,3, &IMXRT_TMR3, &CCM_CCGR6, CCM_CCGR6_QTIMER3(CCM_CCGR_ON), IRQ_QTIMER3, &PulsePositionOutput::isrTimer3 },
    {18,1, &IMXRT_TMR3, &CCM_CCGR6, CCM_CCGR6_QTIMER3(CCM_CCGR_ON), IRQ_QTIMER3, &PulsePositionOutput::isrTimer3 },
    {19,0, &IMXRT_TMR3, &CCM_CCGR6, CCM_CCGR6_QTIMER3(CCM_CCGR_ON), IRQ_QTIMER3, &PulsePositionOutput::isrTimer3 }
};
#define CNT_HARDWARE_LIST (sizeof(hardware)/sizeof(hardware[0]))

PulsePositionOutput * PulsePositionOutput::list[10];


PulsePositionOutput::PulsePositionOutput(void)
{
	pulse_width[0] = TX_MINIMUM_FRAME_CLOCKS;
	for (int i=1; i <= PULSEPOSITION_MAXCHANNELS; i++) {
		pulse_width[i] = TX_DEFAULT_SIGNAL_CLOCKS;
	}
}

PulsePositionOutput::PulsePositionOutput(int polarity)
{
	pulse_width[0] = TX_MINIMUM_FRAME_CLOCKS;
	for (int i=1; i <= PULSEPOSITION_MAXCHANNELS; i++) {
		pulse_width[i] = TX_DEFAULT_SIGNAL_CLOCKS;
	}
	if (polarity == FALLING) {
		outPolarity = 0;
	} else {
		outPolarity = 1;
	}
}

bool PulsePositionOutput::begin(uint8_t txPin)
{
	return begin(txPin, 255);
}

bool PulsePositionOutput::begin(uint8_t txPin, uint8_t framePin)
{
	//uint32_t channel;
	//volatile void *reg;

	Serial.println(txPin);

	for (idx_channel = 0; idx_channel < CNT_HARDWARE_LIST; idx_channel++) {
		if (hardware[idx_channel].pin == txPin) break; 
	}
	if (idx_channel == CNT_HARDWARE_LIST) return false;

	// make sure the appropriate clock gate is enabled.
	*hardware[idx_channel].clock_gate_register |= hardware[idx_channel].clock_gate_mask;

	uint8_t channel =  hardware[idx_channel].channel;
	volatile IMXRT_TMR_CH_t *tmr_ch = &hardware[idx_channel].tmr->CH[channel];

	tmr_ch->CTRL = 0; // stop
	tmr_ch->CNTR = 0;
	tmr_ch->LOAD = 0;

	tmr_ch->COMP1 = 200;  // first time
	state = 0;
	tmr_ch->CMPLD1 = TX_PULSE_WIDTH_CLOCKS;
	//TMR1_SCTRL0 = TMR_SCTRL_OEN | TMR_SCTRL_OPS to make falling
	if(outPolarity == 0){
	  tmr_ch->SCTRL = TMR_SCTRL_OEN | TMR_SCTRL_OPS;
	} else {
	  tmr_ch->SCTRL = TMR_SCTRL_OEN ;
	}

	tmr_ch->CSCTRL = TMR_CSCTRL_CL1(1);
	attachInterruptVector(hardware[idx_channel].interrupt, hardware[idx_channel].isr);
	tmr_ch->CSCTRL &= ~(TMR_CSCTRL_TCF1);  // clear
	tmr_ch->CSCTRL |= TMR_CSCTRL_TCF1EN;  // enable interrupt
	NVIC_SET_PRIORITY(hardware[idx_channel].interrupt, 32);
	NVIC_ENABLE_IRQ(hardware[idx_channel].interrupt);
	tmr_ch->CTRL =  CTRL_SET;

	
	//prev = 0;
	//write_index = 255;
	//available_flag = false;
	list[idx_channel] = this;
	//channelmask |= (1<<idx_channel);
	
	//set Mux for Tx Pin - all timers on ALT1
	*(portConfigRegister(txPin)) = 1;

	/*
	if (framePin < NUM_DIGITAL_PINS) {
		framePinReg = portOutputRegister(framePin);
		framePinMask = digitalPinToBitMask(framePin);
		pinMode(framePin, OUTPUT);
		FRAME_PIN_SET();
	} else {
		framePinReg = NULL;
	}
	*/
	return true;
}

bool PulsePositionOutput::write(uint8_t channel, float microseconds)
{
  uint32_t i, sum, space, clocks, num_channels;

  if (channel < 1 || channel > PULSEPOSITION_MAXCHANNELS) return false;
  if (microseconds < TX_MINIMUM_SIGNAL || microseconds > TX_MAXIMUM_SIGNAL) return false;
  clocks = microseconds * CLOCKS_PER_MICROSECOND;
  num_channels = total_channels;
  if (channel > num_channels) num_channels = channel;
  sum = clocks;
  for (i = 1; i < channel; i++) sum += pulse_width[i];
  for (i = channel + 1; i <= num_channels; i++) sum += pulse_width[i];
  if (sum < TX_MINIMUM_FRAME_CLOCKS - TX_MINIMUM_SPACE_CLOCKS) {
    space = TX_MINIMUM_FRAME_CLOCKS - sum;
  } else {
    if (framePinReg) {
      space = TX_PULSE_WIDTH_CLOCKS;
    } else {
      space = TX_MINIMUM_SPACE_CLOCKS;
    }
  }
  __disable_irq();
  pulse_width[0] = space;
  pulse_width[channel] = clocks;
  total_channels = num_channels;
  __enable_irq();
  return true;
}


void PulsePositionOutput::isr() 
{
  uint8_t channel = hardware[idx_channel].channel;
  volatile IMXRT_TMR_CH_t *tmr_ch = &hardware[idx_channel].tmr->CH[channel];
  ticks++;

  if (state == 0) {
    // pin was just set high, schedule it to go low
	
   tmr_ch->COMP1 = tmr_ch->CMPLD1 = TX_PULSE_WIDTH_CLOCKS;
    tmr_ch->CTRL =  CTRL_CLEAR;
    state = 1;
  } else {
    // pin just went low
    uint32_t width, channel;
    if (state == 1) {
      channel = current_channel;
      if (channel == 0) {
        total_channels_buffer = total_channels;
        for (uint32_t i = 0; i <= total_channels_buffer; i++) {
          pulse_buffer[i] = pulse_width[i];
        }
      }
      width = pulse_buffer[channel] - TX_PULSE_WIDTH_CLOCKS;
      if (++channel > total_channels_buffer) {
        channel = 0;
      }
      if (framePinReg) {
        //if (channel == 0) {
        if (channel == 1) {
          //TODO      FRAME_PIN_SET();
        } else {
          //TODO     FRAME_PIN_CLEAR();
        }
      }
      current_channel = channel;
    } else {
      width = pulse_remaining;
    }
    if (width <= 60000) {
      tmr_ch->COMP1 = tmr_ch->CMPLD1 = width;
      tmr_ch->CTRL =  CTRL_SET; // set on compare match & interrupt
      state = 0;
    } else {
      tmr_ch->COMP1 =tmr_ch->CMPLD1 = 58000;
      tmr_ch->CTRL =  CTRL_CLEAR; // clear on compare match & interrupt
      pulse_remaining = width - 58000;
      state = 2;
    }
  }

}


void PulsePositionOutput::isrTimer1()
{
	digitalWriteFast(2, HIGH);
	if (list[2] && (IMXRT_TMR1.CH[0].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR1.CH[0].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[2]->isr();}
	if (list[3] && (IMXRT_TMR1.CH[2].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR1.CH[2].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[3]->isr();}
	if (list[4] && (IMXRT_TMR1.CH[1].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR1.CH[1].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[4]->isr();}
	asm volatile ("dsb");  // wait for clear  memory barrier
	digitalWriteFast(2, LOW);
}

void PulsePositionOutput::isrTimer2()
{
	digitalWriteFast(2, HIGH);
	if (list[5] && (IMXRT_TMR2.CH[0].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR2.CH[0].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[5]->isr();}
	asm volatile ("dsb");  // wait for clear  memory barrier
	digitalWriteFast(2, LOW);

}


void PulsePositionOutput::isrTimer3()
{
	digitalWriteFast(2, HIGH);
	if (list[6] && (IMXRT_TMR3.CH[2].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR3.CH[2].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[6]->isr();}
	if (list[7] && (IMXRT_TMR3.CH[3].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR3.CH[3].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[7]->isr();}
	if (list[8] && (IMXRT_TMR3.CH[1].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR3.CH[1].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[8]->isr();}
	if (list[9] && (IMXRT_TMR3.CH[0].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR3.CH[0].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[9]->isr();}
	asm volatile ("dsb");  // wait for clear  memory barrier
	digitalWriteFast(2, LOW);
}

void PulsePositionOutput::isrTimer4()
{
	digitalWriteFast(2, HIGH);
	if (list[0] && (IMXRT_TMR4.CH[1].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR4.CH[1].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[0]->isr();}
	if (list[1] && (IMXRT_TMR4.CH[2].CSCTRL & TMR_CSCTRL_TCF1)) {IMXRT_TMR4.CH[2].CSCTRL &= ~(TMR_CSCTRL_TCF1); list[1]->isr();}
	asm volatile ("dsb");  // wait for clear  memory barrier
	digitalWriteFast(2, LOW);
}

